#!/usr/local/bin/wolframscript

(*
   stream_plots_funcs.m
   --
   Bryan Daniels
   7.3.2017
   --
   Functions for finding fixed points in worm models given SBML files.
   Modified from stream_plots_script.m
*)

(* ::Package:: *)

(* ::Input:: *)
(*#!/usr/local/bin/wolframscript*)


(* ::Input:: *)
<<mathsbml.m

notFound[filename_] := Module[{},Print[filename<>" not found."]; 0.]



(* ::Input:: *)
laserPulse[SBMLobj_,intensity_,delay_,t_]:=intensity Piecewise[{
{0,
t<delay},{(t-delay)/0.1,delay<=t<delay+0.1},
{Exp[-(t-delay-0.1)/CTSN`tau\[UnderBracket]pulse],t>=delay+0.1}
}]/.(SBMLConstants/.SBMLobj)

(* 10.2.2017 *)
(*add CTSN` to the beginning of every parameter except those in the \
 given inputs list*)
addCTSN[SBMLobj_,
        inputs_] := (ToExpression[StringDrop[#, 5] <> "->" <> #]) & /@
Select[SBMLParameters /. SBMLobj,
       Not[MemberQ[inputs, StringDrop[#, 5]]] &];

(* 10.2.2017 *)
(*a version of SBMLODES that keeps specified parameters (inputs) \
 unevaluated*)
(*currently bends over backward to look like the output \
 of SBMLODES, but this might be made more simple by seeing how I \
 actually use SBMLODES*)
SBMLODESwithInputs[SBMLobj_, inputs_] :=
(ToExpression[
              StringReplace[
                            StringTake[#, {10, -2}], "=" -> "=="]] & /@
 Select[(SBMLMetaIDAssociations /. SBMLobj)[[;; , 2]],
        StringStartsQ["rateRule["]]
 ) /. addCTSN[SBMLobj, inputs] /. {speed'[t] -> CTSN`speed'[t],
    speed -> CTSN`speed[t],
    laserIntensity\[UnderBracket]timeVarying ->
    CTSN`laserIntensity\[UnderBracket]timeVarying[t],
    X\[UnderBracket]2'[t] -> CTSN`X\[UnderBracket]2'[t],
    X\[UnderBracket]2 -> CTSN`X\[UnderBracket]2[t],
    X\[UnderBracket]3'[t] -> CTSN`X\[UnderBracket]3'[t],
    X\[UnderBracket]3 -> CTSN`X\[UnderBracket]3[t],
    tau\[UnderBracket]1 -> CTSN`tau\[UnderBracket]1[t],
    tau\[UnderBracket]2 -> CTSN`tau\[UnderBracket]2[t],
    tau\[UnderBracket]3 ->
    CTSN`tau\[UnderBracket]3[t]} /. (SBMLConstants /. SBMLobj)

(* 10.2.2017 *)
(*hack around hidden node being called X3 instead of X2---rename to \
 X2*)
(*but only do this if no X2 currently exists*)
renameX3[ODEs_] :=
If[Count[ODEs, CTSN`X\[UnderBracket]2[t], Infinity] == 0,
   ODEs /. {CTSN`X\[UnderBracket]3[t] -> CTSN`X\[UnderBracket]2[t],
       Derivative[1][CTSN`X\[UnderBracket]3][t] ->
       Derivative[1][CTSN`X\[UnderBracket]2][t]}, ODEs]
renameX3params[params_] := 
 If[Count[params, CTSN`X\[UnderBracket]2\[UnderBracket]init, 
    Infinity] == 0, 
  params /. {CTSN`X\[UnderBracket]3\[UnderBracket]init -> 
     CTSN`X\[UnderBracket]2\[UnderBracket]init}, params]

(* ::Input:: *)
(* 10.2.2017 *)
(*version that works with SBML assignment rules (including fixed sign cases) and pauseState input*)
wormODEs[SBMLobj_, laserInput_: 0, pauseStateInput_: 0] :=
renameX3[First[((SBMLODESwithInputs[
                                    SBMLobj, {"pauseState"}]) /. {CTSN`laserIntensity\
    \[UnderBracket]timeVarying[t] -> laserInput,
    pauseState -> pauseStateInput}) /.
               Solve[(SBMLAssignmentRules /. SBMLobj)] /. (SBMLConstants /.
                                                           SBMLobj)]
         ]

(* ::Input:: *)
wormICs[SBMLobj_] := {CTSN`speed[0] == CTSN`speed\[UnderBracket]init, 
   CTSN`X\[UnderBracket]2[0] == 
    CTSN`X\[UnderBracket]2\[UnderBracket]init} /. (renameX3params[
    SBMLConstants /. SBMLobj])

(* ::Input:: *)
wormRHS[SBMLobj_, laserInput_: 0, pauseInput_: 0] :=
First[{Derivative[1][CTSN`speed][t],
    Derivative[1][CTSN`X\[UnderBracket]2][t]} /.
      Solve[wormODEs[SBMLobj, laserInput,
                     pauseInput], {Derivative[1][CTSN`speed][t],
    Derivative[1][CTSN`X\[UnderBracket]2][t]}]] /. {CTSN`speed[t] ->
        CTSN`speed, CTSN`X\[UnderBracket]2[t] -> CTSN`X\[UnderBracket]2}

wormRHS1D[SBMLobj_, laserInput_: 0, pauseInput_: 0] := 
 First[{Derivative[1][CTSN`speed][t], 0} /. 
    Solve[wormODEs[SBMLobj, laserInput, 
      pauseInput], {Derivative[1][CTSN`speed][t]}]] /. {CTSN`speed[
     t] -> CTSN`speed}

X2max[SBMLobj_, laserInput_: 0, pauseInput_: 0] :=
 With[{traj = NDSolve[Join[
      wormODEs[SBMLobj, laserPulse[SBMLobj, laserInput, 10, t], 
       pauseInput], wormICs[SBMLobj]],
     {CTSN`speed[t], CTSN`X\[UnderBracket]2[t]}, {t, 0, 200}]},
  With[{tinit = Table[tmax, {tmax, 10, 15, 0.1}][[First[Ordering[
         Flatten[
          Table[Evaluate[CTSN`X\[UnderBracket]2[t] /. traj], {t, 10, 
            15, 0.1}]], -1]]]]},
   FindMaxValue[
    First[Evaluate[{CTSN`X\[UnderBracket]2[t], 10 < t < 15} /. 
       traj]], {t, tinit}]
   ]
  ]

X2min[SBMLobj_, laserInput_: 0, pauseInput_: 0] :=
 With[{traj = NDSolve[Join[
      wormODEs[SBMLobj, laserPulse[SBMLobj, laserInput, 10, t], 
       pauseInput], wormICs[SBMLobj]],
     {CTSN`speed[t], CTSN`X\[UnderBracket]2[t]}, {t, 0, 200}]},
  With[{tinit = Table[tmin, {tmin, 10, 15, 0.1}][[First[Ordering[
         Flatten[
          Table[Evaluate[CTSN`X\[UnderBracket]2[t] /. traj], {t, 10, 
            15, 0.1}]], 1]]]]},
   FindMinValue[
    First[Evaluate[{CTSN`X\[UnderBracket]2[t], 10 < t < 15} /. 
       traj]], {t, tinit}]
   ]
  ]

(*change Transparent to Thin or Thick to also show unstable fixed \
points*)
fixedPointPlot2D[SBMLobj_, laser_, pointSize_: 0.05] :=
 With[{f = fixedPoints[SBMLobj, laser]},
  ListPlot[Map[{#} &, Transpose[f][[1]]], 
   PlotMarkers -> 
    Map[If[#, {Graphics[{Thin, Darker[Darker[Green]], Circle[]}],
        pointSize}, {Graphics[{Transparent, Darker[Darker[Green]], 
          Circle[]}], pointSize}] &, Transpose[f][[2]]],
   PlotRange -> All
   ]
  ]




(* ************************************************************************* *)
(* 2.8.2017 solve for fixed points at zero laser input and write to csv file *)

dynamicVars[ODEs_] := Map[First[Head[First[#]]][t] &, ODEs]

rhs[ODEs_] := 
 First[D[dynamicVars[ODEs], t] /. Solve[ODEs, D[dynamicVars[ODEs], t]]]

(*wrt dynamical variables and laser*)

JacWorm[ODEs_] := Grad[rhs[ODEs], Join[dynamicVars[ODEs], {laser[s]}]]

JacWorm0[ODEs_] :=
 Grad[rhs[ODEs], dynamicVars[ODEs]]

Jac0Function[SBMLobj_, sVal_?NumericQ, curve_] := 
 First[Evaluate[
   JacWorm0[wormODEs[SBMLobj]] /. {x_[t] -> x[t][s]} /. 
     curve /. {s -> sVal}]]

(*Used to find an initial fixed point in fixedPointCurveLaser*)
findFixedPoint[ODEs_, startMin_: - 100, startMax_: 100] :=
 dynamicVars[ODEs] /. FindRoot[rhs[ODEs] == 0,
   {#, startMin, startMax} & /@ dynamicVars[ODEs]]

curveDomain[curve_] := (laser[s] /. curve)[[1]][[0]]["Domain"][[1]]

detFunction[SBMLobj_, sVal_?NumericQ, curve_] := 
 Det[Jac0Function[SBMLobj, sVal, curve]]

(*from http://mathematica.stackexchange.com/questions/10640/find-zero-\
crossing-in-a-list*)
zeroCrossings[l_List] := 
 Module[{t, u, v}, 
  t = {Sign[l], Range[Length[l]]} // Transpose;(*List of-1,0,1 only*)
  u = Select[t, First[#] != 0 &];(*Ignore zeros*)
  v = SplitBy[u, First];(*Group into runs of+and-
  values*){Most[Max[#[[All, 2]]] & /@ v], 
    Rest[Min[#[[All, 2]]] & /@ v]} // Transpose]

findBifurcation[SBMLobj_, curve_, smin_, smax_] := 
 s /. FindRoot[detFunction[SBMLobj, s, curve], {s, smin, smax}]

(*Note: If numsteps is too small here, some bifurcations could be \
missed.  There may be a better way to do this.*)
sBifurcations[SBMLobj_, curve_, numsteps_: 100] := With[{
   smin = curveDomain[curve][[1]],
   smax = curveDomain[curve][[2]]},
  Module[{sList, signList, crossingIndices},
   sList = Table[s, {s, smin, smax, (smax - smin)/numsteps}];
   signList = Map[detFunction[SBMLobj, #, curve] &, sList];
   crossingIndices = zeroCrossings[signList];
   Map[findBifurcation[SBMLobj, curve, sList[[#[[1]]]], 
      sList[[#[[2]]]]] &, crossingIndices]
   ]]

intervals[lst_] := {lst[[;; -2]], lst[[2 ;;]]}

(*calculate curve of fixed points as laser intensity is allowed to \
vary*)
(*abort if initial fixed point can't be found*)
fixedPointCurveLaser[SBMLobj_, laser0_: 0, smin_: - 1000, 
  smax_: + 1000] :=
 With[{ODEs = wormODEs[SBMLobj, laser[s]]},
  With[{
    vi = Join[
      Table[xV[i][s], {i, Length[dynamicVars[ODEs]]}], {laserV[s]}],
    yi = Join[Map[#[s] &, dynamicVars[ODEs]], {laser[s]}],
    yi0 = Join[Map[#[0] &, dynamicVars[ODEs]], {laser[0]}]},
   NDSolve[{
     D[yi, s] == vi,
     ((JacWorm[ODEs].vi) /. {x_[t] -> x[t][s]}) == 
      Table[0, Length[dynamicVars[ODEs]]],
     Plus @@ (vi^2) == 1,
     yi0 == 
      Join[Check[findFixedPoint[ODEs /. {laser[s] -> laser0}], 
        Abort[]], {laser0}]},
    Join[yi, vi], {s, smin, smax}]
   ]]


(*2.8.2017 version for arbitrary number of dynamical variables*)
singleFixedPointVsLaser[fpCurve_, vars_, laserVal_, smin_, smax_] := 
 First[Evaluate[(#[s] & /@ vars) /. fpCurve] /. 
   FindRoot[
    Evaluate[(laser[s] /. fpCurve) == laserVal], {s, (smin + smax)/2, 
     smin, smax}, PrecisionGoal -> 6]]

(*returns list of fixed points; each fixed point consists of its \
location and a Boolean of whether it is stable (old version as of \
2/8/2017; works only for 2D; assumes at least one direction is \
stable?)*)
fixedPoints2D[SBMLobj_, laser_] :=
 With[{curve = fixedPointCurveLaser[SBMLobj]},
  With[
   {smin = curveDomain[curve][[1]],
    smax = curveDomain[curve][[2]]}, 
   Select[MapThread[{Check[
        singleFixedPointVsLaserWithX2[curve, 
         laser, #1, #2], \[Infinity]], 
       Det[Jac0Function[SBMLobj, (#1 + #2)/2, curve]] > 0} &,
     intervals[Join[{smin}, sBifurcations[SBMLobj, curve], {smax}]]
     ], Not[SameQ[#[[1]], \[Infinity]]] &
    ]
   ]
  ]

(*returns list of fixed points; each fixed point consists of its \
location and a Boolean of whether it is stable*)
(*return empty list on error*)
fixedPoints[SBMLobj_, laser_] :=
 CheckAbort[
  With[{curve = fixedPointCurveLaser[SBMLobj]},
   With[
    {smin = curveDomain[curve][[1]],
     smax = curveDomain[curve][[2]],
     vars = dynamicVars[wormODEs[SBMLobj]]},
    Select[
     MapThread[{
        Check[
         singleFixedPointVsLaser[curve, vars, 
          laser, #1, #2], \[Infinity]], 
        AllTrue[Eigenvalues[
          Jac0Function[SBMLobj, (#1 + #2)/2, curve]], Negative]
        } &,
      intervals[Join[{smin}, sBifurcations[SBMLobj, curve], {smax}]]
      ], Not[SameQ[#[[1]], \[Infinity]]] &
     ]
    ]
   ], {}]



