# clean_ipynb.py
#
# Bryan Daniels
# 8.1.2017
#
# Remove excessively lengthy bits of jupyter notebooks that prevent them
# from being loaded.
#

import json
import sys

def clean(filename,maxsize=1000,newfilename=None):
    """
    Remove any 'outputs' item with length > maxsize.
    
    newfilename (None)              : If None, default to filename[:-6]+'_clean.ipynb'
    """

    if newfilename is None:
        newfilename = filename[:-6]+'_clean.ipynb'

    # read in original file
    with open(filename) as datafile:
        data = json.load(datafile)

    # remove overly large 'outputs' items
    for cell in data['cells']:
        if cell.has_key('outputs'):
            if len(cell['outputs']) > maxsize:
                cell['outputs'] = []

    # write new file
    with open(newfilename, 'w') as outfile:
        json.dump(data, outfile)



if __name__ == '__main__':

    if len(sys.argv) < 2 or len(sys.argv) > 2:
        print "Usage: python clean_ipynb.py notebook_to_clean.ipynb"
        exit()
    else:
        filename = sys.argv[1]

    clean(filename)


